.wf-panel .launcher image {
    transition: 500ms linear;
    -gtk-icon-transform: scale(0.8);
}

.wf-panel .launcher:hover image {
    transition: 500ms linear;
    -gtk-icon-transform: scale(1.2);
}